/* -------------------------------------------- */
import { HeritiersUtility } from "./heritiers-utility.js";
import { HeritiersRollDialog } from "./heritiers-roll-dialog.js";

/* -------------------------------------------- */
const __degatsBonus = [-2, -2, -1, -1, 0, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 8, 8, 9, 9, 10, 10]
const __vitesseBonus = [-2, -2, -1, -1, 0, 0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6, 7, 7, 8]

/* -------------------------------------------- */
/**
 * Extend the base Actor entity by defining a custom roll data structure which is ideal for the Simple system.
 * @extends {Actor}
 */
export class HeritiersActor extends Actor {

  /* -------------------------------------------- */
  /**
   * Override the create() function to provide additional SoS functionality.
   *
   * This overrided create() function adds initial items 
   * Namely: Basic skills, money, 
   *
   * @param {Object} data        Barebones actor data which this function adds onto.
   * @param {Object} options     (Unused) Additional options which customize the creation workflow.
   *
   */

  static async create(data, options) {

    // Case of compendium global import
    if (data instanceof Array) {
      return super.create(data, options);
    }
    // If the created actor has items (only applicable to duplicated actors) bypass the new actor creation logic
    if (data.items) {
      let actor = super.create(data, options);
      return actor;
    }

    if (data.type == 'personnage') {
      const skills = await HeritiersUtility.loadCompendium("fvtt-les-heritiers.competences")
      data.items = []
      for (let skill of skills) {
        if (skill.system.categorie == "utile") {
          data.items.push(skill.toObject())
        }
      }
    }
    if (data.type == 'pnj') {
    }

    return super.create(data, options);
  }

  /* -------------------------------------------- */
  prepareArme(arme) {
    arme = duplicate(arme)
    arme.system.isMelee = HeritiersUtility.isArmeMelee(arme)
    return arme
  }

  /* -------------------------------------------- */
  getWeapons() {
    let armes = []
    for (let arme of this.items) {
      if (arme.type == "arme") {
        armes.push(this.prepareArme(arme))
      }
    }
    return armes
  }
  /* -------------------------------------------- */
  getMonnaies() {
    return this.items.filter(it => it.type == "monnaie")
  }

  /* -----------------------  --------------------- */
  addMember(actorId) {
    let members = duplicate(this.system.members)
    members.push({ id: actorId })
    this.update({ 'system.members': members })
  }
  async removeMember(actorId) {
    let members = this.system.members.filter(it => it.id != actorId)
    this.update({ 'system.members': members })
  }

  /* -----------------------  --------------------- */
  getItemSorted(types) {
    let items = this.items.filter(item => types.includes(item.type)) || []
    HeritiersUtility.sortArrayObjectsByName(items)
    return items
  }
  getEquipments() {
    return this.getItemSorted(["equipement", "accessoire"])
  }
  getAvantages() {
    return this.getItemSorted(["avantage"])
  }
  getDesavantages() {
    return this.getItemSorted(["desavantage"])
  }
  getMonnaies() {
    return this.getItemSorted(["monnaie"])
  }
  getArmors() {
    return this.getItemSorted(["protection"])
  }
  getTalents() {
    return this.getItemSorted(["talent"])
  }
  getContacts() {
    return this.getItemSorted(["contact"])
  }
  getAtouts() {
    return this.getItemSorted(["atoutfeerique"])
  }
  getCapacites() {
    return this.getItemSorted(["capacitenaturelle"])
  }
  getFee() {
    return this.items.find(item => item.type == "fee")
  }
  getProfils() {
    return this.getItemSorted(["profil"])
  }
  getPouvoirs() {
    return this.getItemSorted(["pouvoir"])
  }
  /* -------------------------------------------- */
  getSkills() {
    let comp = []
    for (let item of this.items) {
      item = duplicate(item)
      if (item.type == "competence") {
        comp.push(item)
      }
    }
    return HeritiersUtility.sortByName(comp)
  }

  /* -------------------------------------------- */
  prepareUtileSkill(item) {
    let specList = []
    if (item && item.system.categorie && item.system.categorie == "utile") {
      for (let spec of item.system.specialites) {
        specList.push(spec.name)
      }
    }
    item.nbSpec = specList.length
    item.specList = specList.toString()
  }

  /* -------------------------------------------- */
  organizeUtileSkills(kind = "mental") {
    let comp = {}
    for (let key in game.system.lesheritiers.config.competenceProfil) {
      if (game.system.lesheritiers.config.competenceProfil[key].kind == kind)
        comp[key] = { skills: [], niveau: this.system.competences[key].niveau }
    }
    for (let item of this.items) {
      if (item.type == "competence") {
        if (item.system.categorie == "utile" && comp[item.system.profil]) {
          this.prepareUtileSkill(item)
          comp[item.system.profil].skills.push(item)
        }
      }
    }

    for (let key in comp) {
      HeritiersUtility.sortArrayObjectsByName(comp[key].skills)
    }
    return Object.fromEntries(Object.entries(comp).sort())
  }
  /* -------------------------------------------- */
  organizeContacts() {
    let contactList = {}
    for (let item of this.items) {
      if (item.type == "contact") {
        let c = contactList[item.system.contacttype] || { label: game.system.lesheritiers.config.contactType[item.system.contacttype], list: [] }
        c.list.push(item)
        contactList[item.system.contacttype] = c
      }
    }
    for (let key in contactList) {
      HeritiersUtility.sortArrayObjectsByName(contactList[key].list)
    }
    return contactList
  }

  /* -------------------------------------------- */
  organizeFutileSkills() {
    let comp = []
    for (let item of this.items) {
      if (item.type == "competence") {
        if (item.system.categorie == "futile") {
          comp.push(item)
        }
      }
    }
    HeritiersUtility.sortArrayObjectsByName(comp)
    return HeritiersUtility.sortByName(comp)
  }

  /* -------------------------------------------- */
  getDefenseBase() {
    return Math.max(this.system.attributs.tre.value, this.system.attributs.pui.value)
  }

  /* -------------------------------------------- */
  getVitesseBase() {
    return 5 + __vitesseBonus[this.system.attributs.adr.value]
  }
  /* -------------------------------------------- */
  getProtection() {
    let equipProtection = 0
    for (let armor in this.items) {
      if (armor.type == "protection" && armor.system.equipped) {
        equipProtection += Number(armor.system.protection)
      }
    }
    if (equipProtection < 4) {
      return 4 + equipProtection // Cas des boucliers + sans armure
    }
    return equipProtection // Uniquement la protection des armures + boucliers
  }

  /* -------------------------------------------- */
  getCombatValues() {
    let combat = {
    }
    return combat
  }

  /* -------------------------------------------- */
  prepareBaseData() {
  }

  /* -------------------------------------------- */
  async prepareData() {
    super.prepareData();
  }

  /* -------------------------------------------- */
  prepareDerivedData() {

    if (this.type == 'personnage') {
    }

    super.prepareDerivedData()
  }

  /* -------------------------------------------- */
  _preUpdate(changed, options, user) {

    super._preUpdate(changed, options, user);
  }

  /* -------------------------------------------- */
  getItemById(id) {
    let item = this.items.find(item => item.id == id);
    if (item) {
      item = duplicate(item)
    }
    return item;
  }

  /* -------------------------------------------- */
  async equipItem(itemId) {
    let item = this.items.find(item => item.id == itemId)
    if (item && item.system) {
      let update = { _id: item.id, "system.equipped": !item.system.equipped }
      await this.updateEmbeddedDocuments('Item', [update]); // Updates one EmbeddedEntity
    }
  }

  /* -------------------------------------------- */
  editItemField(itemId, itemType, itemField, dataType, value) {
    let item = this.items.find(item => item.id == itemId)
    if (item) {
      console.log("Item ", item, itemField, dataType, value)
      if (dataType) {
        if (dataType.toLowerCase() == "number") {
          value = Number(value)
        } else {
          value = String(value)
        }
      }
      let update = { _id: item.id, [`system.${itemField}`]: value };
      this.updateEmbeddedDocuments("Item", [update])
    }
  }

  /* -------------------------------------------- */
  getPvMalus() {
    if (this.system.pv.value > 0) {
      if (this.system.pv.value < this.system.pv.max / 2) {
        return { name: "Santé", value: -1 }
      }
      if (this.system.pv.value < 5) {
        return { name: "Santé", value: -2 }
      }
      return { name: "Santé", value: 0 }
    }
    return { name: "Moribond(e)", value: -50 }
  }

  /* -------------------------------------------- */
  compareName(a, b) {
    if (a.name < b.name) {
      return -1;
    }
    if (a.name > b.name) {
      return 1;
    }
    return 0;
  }

  /* -------------------------------------------- */
  getCarac(attrKey) {
    return duplicate(this.system.caracteristiques)
  }

  /* -------------------------------------------- */
  getBonusDegats() {
    return 0;
  }

  /* -------------------------------------------- */
  async equipGear(equipmentId) {
    let item = this.items.find(item => item.id == equipmentId);
    if (item && item.system.data) {
      let update = { _id: item.id, "system.equipped": !item.system.equipped };
      await this.updateEmbeddedDocuments('Item', [update]); // Updates one EmbeddedEntity
    }
  }

  /* -------------------------------------------- */
  getSubActors() {
    let subActors = [];
    for (let id of this.system.subactors) {
      subActors.push(duplicate(game.actors.get(id)));
    }
    return subActors;
  }
  /* -------------------------------------------- */
  async addSubActor(subActorId) {
    let subActors = duplicate(this.system.subactors);
    subActors.push(subActorId);
    await this.update({ 'system.subactors': subActors });
  }
  /* -------------------------------------------- */
  async delSubActor(subActorId) {
    let newArray = [];
    for (let id of this.system.subactors) {
      if (id != subActorId) {
        newArray.push(id);
      }
    }
    await this.update({ 'system.subactors': newArray });
  }

  /* -------------------------------------------- */
  getTotalAdversite() {
    return this.system.adversite.bleue + this.system.adversite.rouge + this.system.adversite.noire
  }

  /* -------------------------------------------- */
  async incDecAdversite(adv, incDec = 0) {
    let adversite = duplicate(this.system.adversite)
    adversite[adv] += Number(incDec)
    adversite[adv] = Math.max(adversite[adv], 0)
    this.update({ 'system.adversite': adversite })
  }
  /* -------------------------------------------- */
  async incDecQuantity(objetId, incDec = 0) {
    let objetQ = this.items.get(objetId)
    if (objetQ) {
      let newQ = objetQ.system.quantite + incDec
      newQ = Math.max(newQ, 0)
      const updated = await this.updateEmbeddedDocuments('Item', [{ _id: objetQ.id, 'system.quantite': newQ }]); // pdates one EmbeddedEntity
    }
  }

  /* -------------------------------------------- */
  computeRichesse() {
    let valueSC = 0
    for (let monnaie of this.items) {
      if (monnaie.type == "monnaie") {
        valueSC += Number(monnaie.system.prixsc) * Number(monnaie.system.quantite)
      }
    }
    return HeritiersUtility.computeMonnaieDetails(valueSC)
  }

  /* -------------------------------------------- */
  computeValeurEquipement() {
    let valueSC = 0
    for (let equip of this.items) {
      if (equip.type == "equipement" || equip.type == "arme" || equip.type == "protection") {
        valueSC += Number(equip.system.prixsc) * Number(equip.system.quantite ?? 1)
        valueSC += (Number(equip.system.prixca) * Number(equip.system.quantite ?? 1)) * 20
        valueSC += (Number(equip.system.prixpo) * Number(equip.system.quantite ?? 1)) * 400
      }
    }
    return HeritiersUtility.computeMonnaieDetails(valueSC)
  }

  /* -------------------------------------------- */
  getCompetence(compId) {
    return this.items.get(compId)
  }

  /* -------------------------------------------- */
  async setPredilectionUsed(compId, predIdx) {
    let comp = this.items.get(compId)
    let pred = duplicate(comp.system.predilections)
    pred[predIdx].used = true
    await this.updateEmbeddedDocuments('Item', [{ _id: compId, 'system.predilections': pred }])
  }

  /* -------------------------------------------- */
  getInitiativeScore() {
    let init = this.getFlag("world", "last-initiative")
    return init || -1
  }

  /* -------------------------------------------- */
  getBestDefenseValue() {
    let defenseList = this.items.filter(item => (item.type == "arme") && item.system.equipped)
    let maxDef = 0
    let bestArme
    for (let arme of defenseList) {
      if (arme.type == "arme") {
        arme = this.prepareArme(arme)
      }
      if (arme.system.totalDefensif > maxDef) {
        maxDef = arme.system.totalDefensif
        bestArme = duplicate(arme)
      }
    }
    return bestArme
  }

  /* -------------------------------------------- */
  searchRelevantTalents(competence) {
    let talents = []

    for (let talent of this.items) {
      if (talent.type == "talent" && talent.system.isautomated && talent.system.automations.length > 0) {
        for (let auto of talent.system.automations) {
          if (auto.eventtype === "prepare-roll") {
            if (auto.competence.toLowerCase() == competence.name.toLowerCase()) {
              talent = duplicate(talent)
              talent.system.bonus = auto.bonus
              talent.system.baCost = auto.baCost
              talents.push(talent)
            }
          }
        }
      }
    }
    return talents
  }
  /* -------------------------------------------- */
  getCurrentParade() {
    if (this.system.statutmasque == "masque") {
      return this.system.combat.parade.masquee
    }
    return this.system.combat.parade.demasquee
  }
  /* -------------------------------------------- */
  getCurrentEsquive() {
    if (this.system.statutmasque == "masque") {
      return this.system.combat.esquive.masquee
    }
    return this.system.combat.esquive.demasquee
  }
  getCurrentResistancePhysique() {
    return this.system.combat.resistancephysique.value
  }
  /* -------------------------------------------- */
  getTricherie() {
    return this.system.rang.tricherie.value
  }
  /* -------------------------------------------- */
  getHeritages() {
    return this.system.rang.heritage.value
  }
  /* -------------------------------------------- */
  incDecTricherie(value) {
    let tricherie = this.system.rang.tricherie
    tricherie.value += value
    tricherie.value = Math.max(tricherie.value, 0)
    tricherie.value = Math.min(tricherie.value, tricherie.max)
    this.update({ 'system.rang.tricherie': tricherie })
  }
  /* -------------------------------------------- */
  getPireCompetence(compName1, compName2) {
    let comp1 = this.items.find( it => it.name == compName1)
    let comp2 = this.items.find( it => it.name == compName2)
    if ( comp1 && comp2 ) {
      if (comp1.system.niveau > comp2.system.niveau) {
        return comp1
      } else {
        return comp2
      }
    }
    return undefined
  }
  /* -------------------------------------------- */
  getCommonRollData(compId = undefined, compName = undefined) {
    let rollData = HeritiersUtility.getBasicRollData()
    rollData.alias = this.name
    rollData.actorImg = this.img
    rollData.actorId = this.id
    rollData.tokenId = this.token?.id
    rollData.img = this.img
    rollData.caracList = this.getCarac()
    rollData.caracKey = "agi"
    rollData.tricherie = this.getTricherie()
    rollData.heritage = this.getHeritages()
    rollData.useTricherie = false
    rollData.useSpecialite = false
    rollData.useHeritage = false
    rollData.rulesMalus.push(this.getPvMalus())

    if (compId) {
      rollData.competence = duplicate(this.items.get(compId) || {})
      this.prepareUtileSkill(rollData.competence)
      rollData.actionImg = rollData.competence?.img
    }
    if (compName) {
      rollData.competence = duplicate(this.items.find(item => item.name.toLowerCase() == compName.toLowerCase()) || {})
      if (rollData.competence && rollData.competence.name) {
        this.prepareUtileSkill(rollData.competence)
        rollData.actionImg = rollData.competence?.img
      } else {
        rollData.competence = undefined
      }
    }
    HeritiersUtility.updateWithTarget(rollData)

    return rollData
  }
  /* -------------------------------------------- */
  async rollInitiative() {
    let rollData = this.getCommonRollData(undefined, "Art de la guerre")
    rollData.mode = "init"
    if (this.system.caracteristiques["san"].value > this.system.caracteristiques["per"].value) {
      rollData.caracKey = "san"
    } else {
      rollData.caracKey = "per"
    }
    rollData.carac = this.system.caracteristiques[rollData.caracKey]
    let rollDialog = await HeritiersRollDialog.create(this, rollData)
    rollDialog.render(true)
  }

  /* -------------------------------------------- */
  async rollCarac(key, isInit = false) {
    let rollData = this.getCommonRollData()
    rollData.mode = "carac"
    rollData.carac = this.system.caracteristiques[key]
    rollData.caracKey = key
    let rollDialog = await HeritiersRollDialog.create(this, rollData)
    rollDialog.render(true)
  }

  /* -------------------------------------------- */
  async rollRang(key, isInit = false) {
    let rollData = this.getCommonRollData()
    rollData.mode = "rang"
    rollData.rang = this.system.rang[key]
    rollData.rangKey = key
    let rollDialog = await HeritiersRollDialog.create(this, rollData)
    rollDialog.render(true)
  }
  /* -------------------------------------------- */
  async rollRootCompetence(compKey) {
    let rollData = this.getCommonRollData()
    rollData.mode = "competence"
    console.log("Compkey", compKey)
    rollData.competence = { name: this.system.competences[compKey].label, system: { niveau: this.system.competences[compKey].niveau } }
    console.log("RollDatra", rollData)
    let rollDialog = await HeritiersRollDialog.create(this, rollData)
    rollDialog.render(true)
  }

  /* -------------------------------------------- */
  async rollCompetence(compId) {
    let rollData = this.getCommonRollData(compId)
    rollData.mode = "competence"
    console.log("RollDatra", rollData)
    let rollDialog = await HeritiersRollDialog.create(this, rollData)
    rollDialog.render(true)
  }

  /* -------------------------------------------- */
  async rollAttaqueArme(armeId) {
    let arme = this.items.get(armeId)
    if (arme) {
      arme = duplicate(arme)
      arme.system.isMelee = HeritiersUtility.isArmeMelee(arme)
      let competenceName = "Tir"
      let key = "prec"
      if (  arme.system.isMelee) {
        competenceName = "Mêlée"
        key = "agi"
      }
      let rollData = this.getCommonRollData(undefined, competenceName)
      rollData.carac = this.system.caracteristiques[key]
      rollData.caracKey = key
      rollData.arme = arme
      rollData.mode = "arme"
      if (rollData.defenderTokenId && arme.system.isMelee ) {
        rollData.cacheDifficulte = true
      }
      console.log(">>>> ARME", rollData)
      let rollDialog = await HeritiersRollDialog.create(this, rollData)
      rollDialog.render(true)
    }
  }

  /* -------------------------------------------- */
  async rollAttaqueBrutaleArme(armeId) {
    let arme = this.items.get(armeId)
    if (arme) {
      let key = "for"
      let competenceName = "Mêlée"
      arme.system.isMelee = HeritiersUtility.isArmeMelee(arme)
      let rollData = this.getCommonRollData(undefined, competenceName)
      rollData.carac = this.system.caracteristiques[key]
      rollData.caracKey = key
      rollData.arme = duplicate(arme)
      rollData.mode = "attaquebrutale"
      rollData.rulesMalus.push({ name: "Attaque brutale", value: -2 } )
      let rollDialog = await HeritiersRollDialog.create(this, rollData)
      rollDialog.render(true)
    }
  }
  /* -------------------------------------------- */
  async rollAttaqueChargeArme(armeId) {
    let arme = this.items.get(armeId)
    if (arme) {
      let key = "agi"
      arme.system.isMelee = HeritiersUtility.isArmeMelee(arme)
      let pireCompetence = this.getPireCompetence("Mêlée", "Mouvement")
      let rollData = this.getCommonRollData(undefined, pireCompetence.name)
      rollData.carac = this.system.caracteristiques[key]
      rollData.caracKey = key
      rollData.arme = duplicate(arme)
      rollData.mode = "attaquecharge"
      let rollDialog = await HeritiersRollDialog.create(this, rollData)
      rollDialog.render(true)
    }
  }
  
  /* -------------------------------------------- */
  async rollAssomerArme(armeId) {
    let arme = this.items.get(armeId)
    if (arme) {
      let competenceName = "Mêlée"
      //arme.system.isMelee = HeritiersUtility.isArmeMelee(arme)
      let rollData = this.getCommonRollData(undefined, competenceName)
      rollData.carac = this.system.caracteristiques["agi"]
      rollData.caracKey = "agi"
      rollData.arme = duplicate(arme)
      rollData.mode = "assommer"
      if (rollData.defenderTokenId) {
        rollData.cacheDifficulte = true
      }
      let rollDialog = await HeritiersRollDialog.create(this, rollData)
      rollDialog.render(true)
    }
  }

  /* -------------------------------------------- */
  async rollPouvoir(pouvoirId) {
    let pouvoir = this.items.get(pouvoirId)
    if (pouvoir) {
      let rollData = this.getCommonRollData(undefined, undefined)
      if (pouvoir.system.feeriemasque != "autre") {
        rollData.pouvoirBase = duplicate(this.system.rang[pouvoir.system.feeriemasque.toLowerCase()])
        rollData.pouvoirBase.label = "Féerie"
        rollData.carac = duplicate(this.system.caracteristiques[pouvoir.system.carac])
        rollData.caracKey = pouvoir.system.carac
      }
      rollData.pouvoir = duplicate(pouvoir)
      rollData.mode = "pouvoir"
      let rollDialog = await HeritiersRollDialog.create(this, rollData)
      rollDialog.render(true)
    }

  }
}
